/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/*************************************************************************
 *** --------------------------------------------------------------------
 ***
 *** Project Name: MosChip
 ***
 *** Module Name: Mos7720
 ***
 *** File: mos7720.c 
 ***		
 ***
 *** File Revision: 1.0 
 ***
 *** Revision Date:  10/05/05 
 ***
 ***
 *** Purpose	  : It gives an interface between USB to Serial 
 ***                and serves as a Serial Driver for the high 
 ***		    level layers /applications.
 ***
 *** Change History:
 ***
 ***
 *** LEGEND	  :
 ***
 *** 
 *** DBG - Code inserted due to as part of debugging
 *** DPRINTK - Debug Print statement
 ***
 *************************************************************************/

/* all file inclusion goes here */

//#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/serial.h>
#include <linux/ioctl.h>
#include <asm/uaccess.h>
#include <linux/usb.h>


#include "mos7720_usbserial.h"
#define KERNEL_2_6		1

#include "mos7720.h"            /* MCS7720 Defines    */
#include "mos7720_16C50.h"		/* 16C50 UART defines */

/* all defines goes here */

/*
 * Debug related defines 
 */

/* 1: Enables the debugging -- 0: Disable the debugging */

#define MOS_DEBUG	0

#if MOS_DEBUG
	static int debug = 0;
	#define DPRINTK(fmt, args...) printk( "%s: " fmt, __FUNCTION__ , ## args)

#else
	static int debug = 0;
	#define DPRINTK(fmt, args...)

#endif



/*
 * Version Information
 */
#define DRIVER_VERSION "1.0.0.4F"
#define DRIVER_DESC "Moschip USB Serial Adapter"

/*
 * Defines used for sending commands to port 
 */

#define WAIT_FOR_EVER   (HZ * 0 ) /* timeout urb is wait for ever*/
#define MOS_WDR_TIMEOUT (HZ * 5 ) /* default urb timeout */
                                                                                                                             
#define MOS_PORT1       0x0200
#define MOS_PORT2       0x0300
#define MOS_VENREG      0x0000
#define MOS_MAX_PORT	0x02                                                                                                                            

#define MCS_RD_RTYPE 0xC0
#define MCS_WR_RTYPE 0x40
 
#define MOS_WRITE       0x0E
#define MOS_READ        0x0D



#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)

static struct usb_serial_driver moschip7720_2port_device = {
        .driver                 = {
                                        .owner  = THIS_MODULE,
                                        .name   = "mos7720",
                                },
        .description            = DRIVER_DESC,
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
static struct usb_serial_device_type moschip7720_2port_device = {
	.owner			= THIS_MODULE,
	.name			= "Moschip USB Serial Adapter",
	.short_name		= "Moschip7720",
#endif
	.id_table		= moschip_port_id_table,
	.num_interrupt_in	= 1,
	.num_bulk_in		= 2,
	.num_bulk_out		= 2,
	.num_ports		= 2,
	.open			= mos7720_open,
	.close			= mos7720_close,
	.throttle		= mos7720_throttle,
	.unthrottle		= mos7720_unthrottle,
	.tiocmget               = mos7720_tiocmget,
        .tiocmset               = mos7720_tiocmset,
	.attach			= mos7720_startup,
	.shutdown		= mos7720_shutdown,
	.ioctl			= mos7720_ioctl,
	.set_termios		= mos7720_set_termios,
	.write			= mos7720_write,
	.write_room		= mos7720_write_room,
	.chars_in_buffer	= mos7720_chars_in_buffer,
	.break_ctl		= mos7720_break,
	.read_bulk_callback	= mos7720_bulk_in_callback, 
	.read_int_callback      = mos7720_interrupt_callback,
};

static struct usb_driver io_driver = {
	.name =		"moschip7720",
	.probe =	usb_serial_probe,
	.disconnect =	usb_serial_disconnect,
	.id_table =	id_table_combined,
};

/************************************************************************/
/*             I N T E R F A C E   F U N C T I O N S			*/
/************************************************************************/

static inline void set_serial_private(struct usb_serial *serial, struct moschip_serial *data)
{
		usb_set_serial_data(serial, (void *)data );
}

static inline struct moschip_serial * get_serial_private(struct usb_serial *serial)
{
		return (struct moschip_serial*) usb_get_serial_data(serial);
}

static inline void set_port_private(struct usb_serial_port *port, struct moschip_port *data)
{
		usb_set_serial_port_data(port, (void*)data );
}

static inline struct moschip_port * get_port_private(struct usb_serial_port *port)
{
	return (struct moschip_port*) usb_get_serial_port_data(port);
}






/************************************************************************/
/*            U S B  C A L L B A C K   F U N C T I O N S                */
/************************************************************************/


/*****************************************************************************
 * mos7720_interrupt_callback
 *	this is the callback function for when we have received data on the 
 *	interrupt endpoint.
 * Input : 1 Input
 *			pointer to the URB packet,
 *
 *****************************************************************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
static void mos7720_interrupt_callback (struct urb *urb)
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
static void mos7720_interrupt_callback (struct urb *urb,struct pt_regs *regs)
#endif
{
       int result;
        int length ;
        struct moschip_port   *mos7720_port;
        struct moschip_serial *mos7720_serial;
        struct usb_serial *serial;
        __u16 Data;
        unsigned char *data;
        __u8 sp[5],st;
        int i;
        __u16 wval;
       // printk("in the function mos7720_interrupt_callback Length %d, Data %x \n",urb->actual_length,(unsigned int)urb->transfer_buffer);
        DPRINTK("%s"," : Entering\n");
        if(!urb)
        {
                DPRINTK("%s","Invalid Pointer !!!!:\n");
                return;
        }
                                                                                                                                                                                                         
        switch (urb->status)
        {
                case 0:
                        /* success */
                        break;
                case -ECONNRESET:
                case -ENOENT:
                case -ESHUTDOWN:
                        /* this urb is terminated, clean up */
                        dbg("%s - urb shutting down with status: %d", __FUNCTION__, urb->status);
                        return;
                default:
                        dbg("%s - nonzero urb status received: %d", __FUNCTION__, urb->status);
                        goto exit;
        }
                                                                                                                                                                                                         
                                                                                                                                                                                                         
        length = urb->actual_length;
        data = urb->transfer_buffer;
                                                                                                                                                                                                         
        mos7720_serial= (struct moschip_serial *)urb->context;
        //serial  = mos7720_get_usb_serial(port,__FUNCTION__);
        serial = mos7720_serial->serial;
                                                                                                                                                                                                         
	 /* Moschip get 4 bytes
         * Byte 1 IIR Port 1 (port.number is 0)
         * Byte 2 IIR Port 2 (port.number is 1)
         * Byte 3 --------------
         * Byte 4 FIFO status for both */

                                                                                                                                                                                                      
        if(length && length>4)
        {
                DPRINTK("%s \n","Wrong data !!!");
                return;
        }
          

        sp[0]=(__u8)data[0];
        sp[1]=(__u8)data[1];
        st=(__u8)data[3];
      //printk("%s data is sp1:%x sp2:%x status:%x\n",__FUNCTION__,sp[0],sp[1],status);
        for(i=0;i<serial->num_ports;i++)
        {
         mos7720_port = get_port_private(serial->port[i]);
         wval = (((__u16)serial->port[i]->number - (__u16)(serial->minor))+1)<<8;
        //printk("%s wval is %x\n",__FUNCTION__,wval);
        if(mos7720_port->open != FALSE)
        {       if(sp[i] & 0x01)
                {
                        DPRINTK("SP%d No Interrupt !!!\n",i);
                }
                else
                {
                        switch(sp[i] & 0x0f)
                        {
                        case SERIAL_IIR_RLS:
                            
                            DPRINTK("Serial Port %d: Receiver status error or ",i);
                            DPRINTK("address bit detected in 9-bit mode\n");
                            mos7720_port->MsrLsr=1;
                            mos7720_get_reg(mos7720_port,wval,LINE_STATUS_REGISTER,&Data);
                            break;
                        case SERIAL_IIR_MS:
                             DPRINTK("Serial Port %d: Modem status change\n",i);
                             mos7720_port->MsrLsr=0;
                             mos7720_get_reg(mos7720_port,wval, MODEM_STATUS_REGISTER, &Data);
                             break;
                        }
                }
        }
        }
	
exit:
        result = usb_submit_urb (urb, GFP_ATOMIC);
        if (result)
        {
                dev_err(&urb->dev->dev, "%s - Error %d submitting interrupt urb\n", __FUNCTION__, result);
        }
                                                                                                                                                                                                         
        return;

}

static int mos7720_get_reg(struct moschip_port *mcs,__u16 Wval, __u16 reg, __u16 * val)
{
        struct usb_device *dev = mcs->port->serial->dev;
        struct usb_ctrlrequest *dr=NULL;
        unsigned char  *buffer=NULL;
        int ret=0;
        buffer= (__u8 *)mcs->ctrl_buf;
                                                                                                                             
//      dr=(struct usb_ctrlrequest *)(buffer);
        dr=(void *)(buffer + 2);
        dr->bRequest = (__u8)0x0D;//MOS_READ;
        dr->bRequestType = MCS_RD_RTYPE;
        dr->wValue = cpu_to_le16(Wval);//0;
        dr->wIndex = cpu_to_le16(reg);
        dr->wLength = cpu_to_le16(2);
        usb_fill_control_urb(mcs->control_urb,dev,usb_rcvctrlpipe(dev,0),(unsigned char *)dr,buffer,2,mos7720_control_callback,mcs);
        mcs->control_urb->transfer_buffer_length = 2;
        ret=usb_submit_urb(mcs->control_urb,GFP_ATOMIC);
        return ret;
}
#if LINUX_VERSION_CODE == KERNEL_VERSION(2,6,19)
static void mos7720_control_callback(struct urb *urb)
#else
static void mos7720_control_callback(struct urb *urb, struct pt_regs *regs)
#endif
{
        unsigned char *data;
        struct moschip_port *mos7720_port;
        __u8 regval=0x0;
                                                                                                                                               
        if(!urb)
        {
                DPRINTK("%s","Invalid Pointer !!!!:\n");
                return;
        }
                                                                                                                                               
        switch (urb->status)
        {
                case 0:
                        /* success */
                        break;
                case -ECONNRESET:
                case -ENOENT:
                case -ESHUTDOWN:
                        /* this urb is terminated, clean up */
                        dbg("%s - urb shutting down with status: %d", __FUNCTION__, urb->status);                        return;
                default:
                        dbg("%s - nonzero urb status received: %d", __FUNCTION__, urb->status);
                        goto exit;
        }
                                                                                                                                               
                                                                                                                                               
        mos7720_port = (struct moschip_port *)urb->context;
                                                                                                                                               
        DPRINTK("%s urb buffer size is %d\n",__FUNCTION__,urb->actual_length);
        DPRINTK("%s mos7720_port->MsrLsr is %d \n",__FUNCTION__,mos7720_port->MsrLsr);
        data=urb->transfer_buffer;
        regval=(__u8)data[0];
        DPRINTK("%s data is %x\n",__FUNCTION__,regval);
        if(mos7720_port->MsrLsr==0)
                handle_newMsr(mos7720_port,regval);
        else if(mos7720_port->MsrLsr==1)
                handle_newLsr(mos7720_port,regval);
                                                                                                                                               
exit:
        return;
}

int handle_newMsr(struct moschip_port *port,__u8 newMsr)
{
        struct moschip_port *mos7720_port;
        struct  async_icount *icount;
        mos7720_port=port;
        icount = &mos7720_port->icount;
        if (newMsr & (MOS_MSR_DELTA_CTS | MOS_MSR_DELTA_DSR | MOS_MSR_DELTA_RI | MOS_MSR_DELTA_CD)) {
                icount = &mos7720_port->icount;
                                                                                                                             
                /* update input line counters */
                if (newMsr & MOS_MSR_DELTA_CTS) {
                        icount->cts++;
                }
                if (newMsr & MOS_MSR_DELTA_DSR) {
                        icount->dsr++;
                }
                if (newMsr & MOS_MSR_DELTA_CD) {
                        icount->dcd++;
                }
                if (newMsr & MOS_MSR_DELTA_RI) {
                        icount->rng++;
                }
        }
                                                                                                                             
                                                                                                                             
        return 0;
}

int handle_newLsr(struct moschip_port *port,__u8 newLsr)
{
        struct  async_icount *icount;
                                                                                                                             
        dbg("%s - %02x", __FUNCTION__, newLsr);
                                                                                                                             
                                                                                                                             
        if (newLsr & SERIAL_LSR_BI) {
                //
                // Parity and Framing errors only count if they
                // occur exclusive of a break being
                // received.
                //
                newLsr &= (__u8)(SERIAL_LSR_OE | SERIAL_LSR_BI);
        }
                                                                                                                             
                                                                                                                             
        /* update input line counters */
        icount = &port->icount;
        if (newLsr & SERIAL_LSR_BI) {
                icount->brk++;
        }
        if (newLsr & SERIAL_LSR_OE) {
                icount->overrun++;
        }
        if (newLsr & SERIAL_LSR_PE) {
                icount->parity++;
        }
        if (newLsr & SERIAL_LSR_FE) {
                icount->frame++;
        }
                                                                                                                             
                                                                                                                             
        return 0;
}




/*****************************************************************************
 * mos7720_bulk_in_callback
 *	this is the callback function for when we have received data on the 
 *	bulk in endpoint.
 * Input : 1 Input
 *			pointer to the URB packet,
 *
 *****************************************************************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
static void mos7720_bulk_in_callback (struct urb *urb)
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
static void mos7720_bulk_in_callback (struct urb *urb, struct pt_regs *regs)
#endif
{
	int			status;
	unsigned char		*data ;
	struct usb_serial	*serial;
	struct usb_serial_port	*port;
	struct moschip_serial	*mos7720_serial;
	struct moschip_port	*mos7720_port;
	struct tty_struct *tty;
	int i;


	if(!urb)
	{
		DPRINTK("%s","Invalid Pointer !!!!:\n");
		return;
	}

	if (urb->status) 
	{
		DPRINTK("nonzero read bulk status received: %d",urb->status);
//		if(urb->status==84)
		//ThreadState=1;
		return;
	}

	mos7720_port= (struct moschip_port*)urb->context;
	if(!mos7720_port)
	{
		DPRINTK("%s","NULL mos7720_port pointer \n");
		return ;
	}

	port = (struct usb_serial_port *)mos7720_port->port;

	serial = port->serial;
	if(!serial)
	{
		DPRINTK("%s\n","Bad serial pointer ");
		return;
	}

	DPRINTK("%s\n","Entering... \n");

	data = urb->transfer_buffer;
	mos7720_serial = get_serial_private(serial);

	DPRINTK("%s","Entering ........... \n");

	if (urb->actual_length) 
	{

	 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
                // 2.6.17 Block
                tty = mos7720_port->port->tty;
                if (tty)
                {
                        tty_buffer_request_room(tty, urb->actual_length);
                        tty_insert_flip_string(tty, data, urb->actual_length);
                                DPRINTK(" %s \n",data);
                        tty_flip_buffer_push(tty);
                }
                                                                                                                             
        #endif
                                                                                                                             
        #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)


		tty = mos7720_port->port->tty;
		if (tty) 
		{
			for (i = 0; i < urb->actual_length; ++i) 
			{
			/* if we insert more than TTY_FLIPBUF_SIZE characters, we drop them. */
				if(tty->flip.count >= TTY_FLIPBUF_SIZE) 
				{
						tty_flip_buffer_push(tty);
				}
			/* this doesn't actually push the data through unless tty->low_latency is set */
				tty_insert_flip_char(tty, data[i], 0);
					DPRINTK(" %c \n",data[i]);
			}
				tty_flip_buffer_push(tty);
		}
	#endif
		mos7720_port->icount.rx += urb->actual_length;
                DPRINTK("mos7720_port->icount.rx is %d:\n",mos7720_port->icount.rx);
 
		
	}

	if(!mos7720_port->read_urb)
	{
		DPRINTK("%s","URB KILLED !!!\n");
		return;
	}

	if(mos7720_port->read_urb->status!=-EINPROGRESS)
	{
		mos7720_port->read_urb->dev = serial->dev;

		status = usb_submit_urb(mos7720_port->read_urb, GFP_ATOMIC);

		if (status) 
		{
			DPRINTK(" usb_submit_urb(read bulk) failed, status = %d", status);
		}
	}
}

/*****************************************************************************
 * mos7720_bulk_out_data_callback
 *	this is the callback function for when we have finished sending serial data
 *	on the bulk out endpoint.
 * Input : 1 Input
 *			pointer to the URB packet,
 *
 *****************************************************************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
static void mos7720_bulk_out_data_callback (struct urb *urb)
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
static void mos7720_bulk_out_data_callback (struct urb *urb, struct pt_regs *regs)
#endif
{
	struct moschip_port *mos7720_port ;
	struct tty_struct *tty;

	if(!urb)
	{
		DPRINTK("%s","Invalid Pointer !!!!:\n");
		return;
	}

	if (urb->status) 
	{
		DPRINTK("nonzero write bulk status received:%d\n", urb->status);
		return;
	}

	mos7720_port = (struct moschip_port *)urb->context;
	if(!mos7720_port)
	{
		DPRINTK("%s","NULL mos7720_port pointer \n");
		return ;
	}


	DPRINTK("%s \n","Entering .........");

	tty = mos7720_port->port->tty;

	if (tty && mos7720_port->open) 
	{
		/* let the tty driver wakeup if it has a special *
		 * write_wakeup function 			 */

		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup) {
			(tty->ldisc.write_wakeup)(tty);
		}

		/* tell the tty driver that something has changed */
		wake_up_interruptible(&tty->write_wait);
	}

	/* Release the Write URB */
	mos7720_port->write_in_progress = FALSE;

//schedule_work(&mos7720_port->port->work);
	
}

/*****************************************************************************
 * SendMosCmd
 *	this function will be used for sending command to device
 * Input : 5 Input
 *			pointer to the serial device,
 *			request type
 *			value
 *			register index
 *			pointer to data/buffer
 *****************************************************************************/


static int SendMosCmd(struct usb_serial *serial, __u8 request, __u16 value, __u16 index, void *data)
{
	int timeout;
	int status;

	__u8 requesttype;
	__u16 size;
	unsigned int Pipe;


    size = 0x00;
    timeout = MOS_WDR_TIMEOUT;
    if(value < MOS_MAX_PORT)
		value = value*0x100+0x200 ;
    else
		value = 0x0000;
        
	if(request==MOS_WRITE)
    {
        request = (__u8)MOS_WRITE;
        requesttype = (__u8)0x40;
        value  = value + (__u16)*((unsigned char *)data);
        data = NULL;
	Pipe = usb_sndctrlpipe(serial->dev, 0);
    }
    else
    {
        request = (__u8)MOS_READ;
        requesttype = (__u8)0xC0;
		size = 0x01;
		Pipe = usb_rcvctrlpipe(serial->dev,0); 
    }
                                                                                                                             
    status = usb_control_msg(           	        \
    serial->dev,                	       		\
    Pipe,           					\
    request,                                    	\
    requesttype,                                    	\
    value,                                          	\
    index,                                          	\
    data,                                           	\
    size,                                           	\
    timeout );
                                                                                                                         
    if(status<0)
    {
        DPRINTK("Command Write failed Value %x index %x\n",value,index);
	DPRINTK("%s\n"," sendmoscmd:leaving...");
	//ThreadState = 1; 
	return status;

    }

        return status;
}



/************************************************************************/
/*       D R I V E R  T T Y  I N T E R F A C E  F U N C T I O N S       */
/************************************************************************/


/*****************************************************************************
 * SerialOpen
 *	this function is called by the tty driver when a port is opened
 *	If successful, we return 0
 *	Otherwise we return a negative error number.
 *****************************************************************************/

static int mos7720_open (struct usb_serial_port *port, struct file * filp)
{
    int response;
    char data;
    int j;
    struct usb_serial *serial;
    struct usb_serial_port *port0;
    struct urb *urb;

    struct moschip_serial *mos7720_serial;
    struct moschip_port *mos7720_port;
                                                                                                                         
	serial = port->serial;


	mos7720_port = get_port_private(port); 

	if (mos7720_port == NULL)
		return -ENODEV;

	port0 = serial->port[0];

	mos7720_serial = get_serial_private(serial);

	if (mos7720_serial == NULL || port0 == NULL) 
	{
		return -ENODEV;
	}
	mos7720_serial->NoOfOpenPorts++;


     usb_clear_halt(serial->dev, port->write_urb->pipe);
     usb_clear_halt(serial->dev, port->read_urb->pipe);
	
		 /* Initialising the write urb pool */
                 for (j = 0; j < NUM_URBS; ++j)
                 {
                        urb = usb_alloc_urb(0,SLAB_ATOMIC);
                        mos7720_port->write_urb_pool[j] = urb;
                                                                                                                             
                        if (urb == NULL)
                        {
                                err("No more urbs???");
                                continue;
                        }
                        
			urb->transfer_buffer = NULL;
                        urb->transfer_buffer = kmalloc (URB_TRANSFER_BUFFER_SIZE, GFP_KERNEL);
                        if (!urb->transfer_buffer)
                        {
                                err("%s-out of memory for urb buffers.", __FUNCTION__);
                                continue;
                        }
                }


/*****************************************************************************
 * Initialize MCS7720 -- Write Init values to corresponding Registers
 *
 * Register Index
 * 1 : IER
 * 2 : FCR
 * 3 : LCR
 * 4 : MCR
 *
 * 0x08 : SP1/2 Control Reg
 *****************************************************************************/

	SendMosCmd(port->serial,MOS_READ,mos7720_port->port->number - mos7720_port->port->serial->minor,LSR, &data);
	DPRINTK("\nSS::%x               LSR:%x\n",(unsigned int)mos7720_port,data);

	DPRINTK("%s", "Check:Sending Command .......... \n");


	 /*
        data = 0x02;
        SendMosCmd(port->serial,MOS_WRITE,MOS_MAX_PORT,0x01, &data);
        data = 0x02;
        SendMosCmd(port->serial,MOS_WRITE,MOS_MAX_PORT,0x02, &data);
        */
        //setting SPregisters
        data=0x02;
        SendMosCmd(port->serial,MOS_WRITE,MOS_MAX_PORT,(__u8)(port->number - port->serial->minor + 1), &data);



	data = 0x00;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x01, &data);
	data = 0x00;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x02, &data);

	data = 0xCF;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x02, &data);
	data = 0x03;
        mos7720_port->shadowLCR  = data;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x03, &data);
	data = 0x0b;
        mos7720_port->shadowMCR  = data;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x04, &data);
	data = 0x0b;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x04, &data);

	data = 0x00;
	SendMosCmd(port->serial,MOS_READ,MOS_MAX_PORT,0x08, &data);
	data = 0x00;
	SendMosCmd(port->serial,MOS_WRITE,MOS_MAX_PORT,0x08, &data);
	
/*	data = 0x00;
	SendMosCmd(port->serial,MOS_READ,MOS_MAX_PORT,port->number - port->serial->minor + 1, &data);
	data = 0x03;
	SendMosCmd(port->serial,MOS_WRITE,MOS_MAX_PORT,port->number - port->serial->minor + 1, &data);
	data = 0x00;
	SendMosCmd(port->serial,MOS_WRITE,MOS_MAX_PORT,port->number - port->serial->minor + 1, &data);
*/
	data = 0x00;
	SendMosCmd(port->serial,MOS_READ,MOS_MAX_PORT,0x08, &data);

	data = data | (port->number - port->serial->minor + 1);
	SendMosCmd(port->serial,MOS_WRITE,MOS_MAX_PORT,0x08, &data);

	data = 0x83;
        mos7720_port->shadowLCR  = data;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x03, &data);
	data = 0x0c;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x00, &data);
	data = 0x00;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x01, &data);
	data = 0x03;
        mos7720_port->shadowLCR  = data;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x03, &data);
	data = 0x0c;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x01, &data);
	data = 0x0c;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x01, &data);


	/* force low_latency on so that our tty_push actually forces *
	 * the data through,otherwise it is scheduled, and with      *
	 * high data rates (like with OHCI) data can get lost.       */
        
	if (port->tty)
		port->tty->low_latency = 1;                                                                                                                     
    	/* see if we've set up our endpoint info yet   *
	 * (can't set it up in mos7720_startup as the  *
	 * structures were not set up at that time.)   */

    if(mos7720_serial->NoOfOpenPorts==1)
    {		
    if (mos7720_serial->interrupt_in_buffer == NULL) 
    {
	DPRINTK("%s \n","Interrupt buffer NULL !!!");

	/* not set up yet, so do it now */

	mos7720_serial->interrupt_in_buffer = serial->port[0]->interrupt_in_buffer;

	mos7720_serial->interrupt_in_endpoint = serial->port[0]->interrupt_in_endpointAddress;
	mos7720_serial->interrupt_read_urb = serial->port[0]->interrupt_in_urb;
	


	DPRINTK("%s \n","To Submit URB !!!");

	/* set up our interrupt urb */

	usb_fill_int_urb(					\
		mos7720_serial->interrupt_read_urb,		\
		serial->dev,					\
		usb_rcvintpipe(serial->dev,			\
		port->interrupt_in_endpointAddress),		\
		port0->interrupt_in_buffer,			\
		mos7720_serial->interrupt_read_urb->transfer_buffer_length,\
		mos7720_interrupt_callback, mos7720_serial,\
		mos7720_serial->interrupt_read_urb->interval );


	/* start interrupt read for this mos7720 this interrupt * 
         * will continue as long as the mos7720 is connected    */

	DPRINTK("%s \n","Submit URB over !!!");
	
	response =usb_submit_urb(mos7720_serial->interrupt_read_urb,GFP_KERNEL);

	if (response) 
	{
	err("%s - Error %d submitting control urb", __FUNCTION__, response);
	}
	
   }
   }

	mos7720_port->bulk_in_buffer    = port->bulk_in_buffer;
    	mos7720_port->bulk_in_endpoint  = port->bulk_in_endpointAddress;
	mos7720_port->read_urb          = port->read_urb;
	mos7720_port->bulk_out_endpoint = port->bulk_out_endpointAddress;

	/* set up our bulk in urb */

	usb_fill_bulk_urb(  
		mos7720_port->read_urb, 				\
		serial->dev,						\
		usb_rcvbulkpipe(serial->dev, port->bulk_in_endpointAddress),\
		port->bulk_in_buffer,					\
		mos7720_port->read_urb->transfer_buffer_length,		\
		mos7720_bulk_in_callback, mos7720_port);	
	

	response = usb_submit_urb (mos7720_port->read_urb,GFP_KERNEL);

	if (response) 
	{
		err("%s - Error %d submitting control urb", __FUNCTION__, response);
	}

	/* initialize our wait queues */
	init_waitqueue_head(&mos7720_port->wait_open);
	init_waitqueue_head(&mos7720_port->wait_chase);
	init_waitqueue_head(&mos7720_port->delta_msr_wait);
	init_waitqueue_head(&mos7720_port->wait_command);

	/* initialize our icount structure */
	memset (&(mos7720_port->icount), 0x00, sizeof(mos7720_port->icount));

	/* initialize our port settings */
	mos7720_port->shadowMCR  = MCR_MASTER_IE; /* Must set to enable ints! */
	mos7720_port->chaseResponsePending = FALSE;                                                                                                
	/* send a open port command */
	mos7720_port->openPending = FALSE;
	mos7720_port->open        = TRUE;                                                                                                                     
	mos7720_port->rxBytesAvail = 0x0;	


	DPRINTK("\n\nusb_serial serial:%x	mos7720_port:%x\nmos7720_serial:%x	usb_serial_port port:%x\n\n",(unsigned int)serial,(unsigned int)mos7720_port,(unsigned int)mos7720_serial,(unsigned int)port);



	return 0;
}

/*****************************************************************************
 * mos7720_close
 *	this function is called by the tty driver when a port is closed
 *****************************************************************************/

static void mos7720_close (struct usb_serial_port *port, struct file * filp)
{
	struct usb_serial *serial;
	struct moschip_serial *mos7720_serial;
	struct moschip_port *mos7720_port;
	char data;
	int j;
	
	DPRINTK("%s\n","mos7720_close:entering...");
		 
	serial = port->serial;
	if (!serial)
	{
		DPRINTK("%s","Serial Paranoia failed \n");
		return;
	}

	mos7720_serial = get_serial_private(serial);
	mos7720_port = get_port_private(port);
	
	if ((mos7720_serial == NULL) || (mos7720_port == NULL))
	{
		return;
	}

	/* flush and block until tx is empty*/
	if (serial->dev) 
		block_until_tx_empty(mos7720_port);


	// Kill The Port's URBs	
	for (j = 0; j < NUM_URBS; ++j)
	#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
                usb_kill_urb (mos7720_port->write_urb_pool[j]);
        #endif
        #if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,10)
                usb_unlink_urb (mos7720_port->write_urb_pool[j]);
        #endif


	   /* Freeing Write URBs*/
	   for (j = 0; j < NUM_URBS; ++j)
 	   {
                if (mos7720_port->write_urb_pool[j])
                {
                    if (mos7720_port->write_urb_pool[j]->transfer_buffer)
                            kfree(mos7720_port->write_urb_pool[j]->transfer_buffer);
                                                                                                                             
                    usb_free_urb (mos7720_port->write_urb_pool[j]);
                }
          }

		/* flush and block until tx is empty*/
	//if (serial->dev)                             // note : moved up
	//	block_until_tx_empty(mos7720_port);

	if (serial->dev) 
	{
		 if (mos7720_port->write_urb)
                {
                        DPRINTK("%s","Shutdown bulk write\n");
        #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
                        usb_kill_urb (mos7720_port->write_urb);
        #endif
        #if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,10)
                        usb_unlink_urb (mos7720_port->write_urb);
        #endif
                }
                                                                                                                                               
                if (mos7720_port->read_urb)
                {
                        DPRINTK("%s","Shutdown bulk read\n");
        #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
                        usb_kill_urb (mos7720_port->read_urb);
        #endif
        #if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,10)
                        usb_unlink_urb (mos7720_port->read_urb);
        #endif
                }
                if((&mos7720_port->control_urb))
                {
                        DPRINTK("%s","Shutdown control read\n");
                //      usb_kill_urb (mos7720_port->control_urb);
                                                                                                                                               
                }
	}
        
//              if(mos7720_port->ctrl_buf != NULL)
  //                      kfree(mos7720_port->ctrl_buf);
	        mos7720_serial->NoOfOpenPorts--;
        	DPRINTK("NoOfOpenPorts in close%d:in port%d\n",mos7720_serial->NoOfOpenPorts,port->number);
	        if(mos7720_serial->NoOfOpenPorts==0)
        	{
                if(mos7720_serial->interrupt_read_urb)
                {
                        DPRINTK("%s","Shutdown interrupt_read_urb\n");
                //      mos7720_serial->interrupt_in_buffer=NULL;
                //      usb_kill_urb (mos7720_serial->interrupt_read_urb);
                }
        
		}
	

	/* While closing port, shutdown all bulk read, write  *
	 * and interrupt read if they exists                  */


	if (mos7720_port->write_urb) 
	{
		/* if this urb had a transfer buffer already (old tx) free it */

		if (mos7720_port->write_urb->transfer_buffer != NULL) 
		{
			kfree(mos7720_port->write_urb->transfer_buffer);
		}
		usb_free_urb(mos7720_port->write_urb);
	}


	data = 0x00;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x04, &data);

	data = 0x00;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x01, &data);

	mos7720_port->open         = FALSE;
	mos7720_port->closePending = FALSE;
	mos7720_port->openPending  = FALSE;

	DPRINTK("%s \n","Leaving ............");
}   


/*****************************************************************************
 * SerialBreak
 *	this function sends a break to the port
 *****************************************************************************/
static void mos7720_break (struct usb_serial_port *port, int break_state)
{
        unsigned char data;
	struct usb_serial *serial;
	struct moschip_serial *mos7720_serial;
	struct moschip_port *mos7720_port;
	
	DPRINTK("%s \n","Entering ...........");

		 
	serial = port->serial;
	if (!serial)
	{
		DPRINTK("%s","Serial Paranoia failed \n");
		return;
	}

	mos7720_serial = get_serial_private(serial);
	mos7720_port = get_port_private(port);
	
	if ((mos7720_serial == NULL) || (mos7720_port == NULL))
	{
		return;
	}
	
	/* flush and chase */
	mos7720_port->chaseResponsePending = TRUE;

	if (serial->dev) 
	{

		/* flush and block until tx is empty*/
		block_until_chase_response(mos7720_port);
	}

        if(break_state == -1)
        {
                data = mos7720_port->shadowLCR | LCR_SET_BREAK;
        }
        else
        {
                data = mos7720_port->shadowLCR & ~LCR_SET_BREAK;
        }

        mos7720_port->shadowLCR  = data;
        SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x03, &data);

	return;
}


/************************************************************************
 *
 * block_until_chase_response
 *
 *	This function will block the close until one of the following:
 *		1. Response to our Chase comes from mos7720
 *		2. A timout of 10 seconds without activity has expired
 *		   (1K of mos7720 data @ 2400 baud ==> 4 sec to empty)
 *
 ************************************************************************/

static void block_until_chase_response(struct moschip_port *mos7720_port)
{
	int timeout = 1*HZ;
	int wait = 10;
	int count ;


	while (1) 
	{
		count = mos7720_chars_in_buffer(mos7720_port->port);
		
		/* Check for Buffer status */
		if(count<=0)
		{
			mos7720_port->chaseResponsePending = FALSE;
			return;
		}

		/* Block the thread for a while */
		interruptible_sleep_on_timeout (&mos7720_port->wait_chase, timeout);
                /* No activity.. count down section */
		wait--;
		if (wait == 0) 
		{
			dbg("%s - TIMEOUT", __FUNCTION__);
			return;
		}
		else 
		{
	                /* Reset timout value back to seconds */	
			wait = 10;
		}
	}

}


/************************************************************************
 *
 * block_until_tx_empty
 *
 *	This function will block the close until one of the following:
 *		1. TX count are 0
 *		2. The mos7720 has stopped
 *		3. A timout of 3 seconds without activity has expired
 *
 ************************************************************************/
static void block_until_tx_empty (struct moschip_port *mos7720_port)
{
	int timeout = HZ/10;
	int wait = 30;
	int count;

	while (1) 
	{

		count = mos7720_chars_in_buffer(mos7720_port->port);

                /* Check for Buffer status */
		if(count<=0)
		{
			return;
		}

                /* Block the thread for a while */
		interruptible_sleep_on_timeout (&mos7720_port->wait_chase, timeout);

                /* No activity.. count down section */
		wait--;
		if (wait == 0) 
		{
			dbg("%s - TIMEOUT", __FUNCTION__);
			return;
		}
		else 
		{
	                /* Reset timout value back to seconds */
			wait = 30;
		}
	}
}

/*****************************************************************************
 * mos7720_write_room
 *	this function is called by the tty driver when it wants to know how many
 *	bytes of data we can accept for a specific port.
 *	If successful, we return the amount of room that we have for this port
 *	Otherwise we return a negative error number.
 *****************************************************************************/

static int mos7720_write_room (struct usb_serial_port *port)
{
    int i;
    int room = 0;
    struct moschip_port *mos7720_port;
	

	DPRINTK("%s \n"," mos7720_write_room:entering ...........");


        mos7720_port = get_port_private(port);
        if (mos7720_port == NULL)
        {
                DPRINTK("%s \n","mos7720_break:leaving ...........");
                return -1;
        }
                                                                                                                             
        for (i = 0; i < NUM_URBS; ++i) 
	{
                if (mos7720_port->write_urb_pool[i]->status != -EINPROGRESS) 
		{
                        room += URB_TRANSFER_BUFFER_SIZE;
                }
        }
       
        dbg("%s - returns %d", __FUNCTION__, room);
        return (room);

}


/*****************************************************************************
 * mos7720_chars_in_buffer
 *	this function is called by the tty driver when it wants to know how many
 *	bytes of data we currently have outstanding in the port (data that has
 *	been written, but hasn't made it out the port yet)
 *	If successful, we return the number of bytes left to be written in the 
 *	system, 
 *	Otherwise we return a negative error number.
 *****************************************************************************/

static int mos7720_chars_in_buffer (struct usb_serial_port *port)
{
	int i;
    	int chars = 0;
 	struct moschip_port *mos7720_port;

	DPRINTK("%s \n"," mos7720_chars_in_buffer:entering ...........");

                                                                                                                             
        mos7720_port = get_port_private(port);
        if (mos7720_port == NULL)
        {
                DPRINTK("%s \n","mos7720_break:leaving ...........");
                return -1;
        }
                                                                                                                       
    	for (i = 0; i < NUM_URBS; ++i) 
    	{
        	if (mos7720_port->write_urb_pool[i]->status == -EINPROGRESS) 
	    	{
                	chars += URB_TRANSFER_BUFFER_SIZE;
            	}
    	}                                                                                                                         
    	dbg("%s - returns %d", __FUNCTION__, chars);
    	return (chars);

}


/*****************************************************************************
 * SerialWrite
 *	this function is called by the tty driver when data should be written to
 *	the port.
 *	If successful, we return the number of bytes written, otherwise we 
 *      return a negative error number.
 *****************************************************************************/

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
static int mos7720_write (struct usb_serial_port *port, const unsigned char *data, int count)
#endif
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,10)
static int mos7720_write (struct usb_serial_port *port, int from_user, const unsigned char *data, int count)
#endif
{
	int status;
	int i;
	int bytes_sent = 0;
	int transfer_size;
	#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
        int from_user=0;
        #endif


	struct moschip_port *mos7720_port;
	struct usb_serial *serial;
	struct moschip_serial *mos7720_serial;
	struct urb    *urb;

	const unsigned char *current_position = data;
	
	DPRINTK("%s \n","entering ...........");


	serial = port->serial;

	mos7720_port = get_port_private(port);
	if(mos7720_port==NULL)
	{
		DPRINTK("%s","mos7720_port is NULL\n");
		return -1;
	}

	mos7720_serial =get_serial_private(serial);
	if(mos7720_serial==NULL)
	{
		DPRINTK("%s","mos7720_serial is NULL \n");
		return -1;
	}

	
        	/* try to find a free urb in the list */
        	urb = NULL;

        	for (i = 0; i < NUM_URBS; ++i) 
		{
                	if (mos7720_port->write_urb_pool[i]->status != -EINPROGRESS) 
			{
                        	urb = mos7720_port->write_urb_pool[i];
				DPRINTK("\nURB:%d",i);
                        	break;
                	}
        	}

	        if (urb == NULL) 
		{
                	dbg("%s - no more free urbs", __FUNCTION__);
                	goto exit;
        	}

        	if (urb->transfer_buffer == NULL) 
		{
          	urb->transfer_buffer = kmalloc (URB_TRANSFER_BUFFER_SIZE, GFP_KERNEL);

                	if (urb->transfer_buffer == NULL) 
			{
                        	err("%s no more kernel memory...", __FUNCTION__);
                        	goto exit;
                	}
        	}                                                                                                                   
        	transfer_size = min (count, URB_TRANSFER_BUFFER_SIZE);

        	if (from_user) 
		{
                	if (copy_from_user (urb->transfer_buffer, current_position, transfer_size)) 	{
                        	bytes_sent = -EFAULT;
	                       	goto exit;
            		}
        	} 
		else 
		{
                	memcpy (urb->transfer_buffer, current_position, transfer_size);
        	}                                                                                                                  
        	//usb_serial_debug_data (__FILE__, __FUNCTION__, transfer_size, urb->transfer_buffer);

		/* fill urb with data and submit  */

           	usb_fill_bulk_urb (urb,
		 mos7720_serial->serial->dev,
		 usb_sndbulkpipe(mos7720_serial->serial->dev,
		 port->bulk_out_endpointAddress),
		 urb->transfer_buffer, 
		 transfer_size,
		 mos7720_bulk_out_data_callback,
		 mos7720_port);


        	/* send it down the pipe */
        	status = usb_submit_urb(urb,GFP_ATOMIC);

        	if (status) 
		{
                	err("%s - usb_submit_urb(write bulk) failed with status = %d", __FUNCTION__, status);
                	bytes_sent = status;
			goto exit;
        	}
	       	bytes_sent = transfer_size;
                mos7720_port->icount.tx += transfer_size;
                DPRINTK("mos7720_port->icount.tx is %d:\n",mos7720_port->icount.tx);
    	 
exit:
	return bytes_sent;

}

/*****************************************************************************
 * SerialThrottle
 *	this function is called by the tty driver when it wants to stop the data
 *	being read from the port.
 *****************************************************************************/

static void mos7720_throttle (struct usb_serial_port *port)
{
	struct moschip_port *mos7720_port;
	struct tty_struct *tty;
	int status;


	DPRINTK("- port %d\n", port->number);

	mos7720_port = get_port_private(port); 

	if (mos7720_port == NULL)
		return;

	if (!mos7720_port->open) 
	{
		DPRINTK("%s\n","port not opened");
		return;
	}

	DPRINTK("%s","Entering .......... \n");

	tty = port->tty;
	if (!tty) 
	{
		dbg ("%s - no tty available", __FUNCTION__);
		return;
	}

	/* if we are implementing XON/XOFF, send the stop character */
	if (I_IXOFF(tty)) 
	{
		unsigned char stop_char = STOP_CHAR(tty);
		#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
                status = mos7720_write (port, &stop_char, 1); //FC4
	        #endif
	        #if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,10)
                status = mos7720_write (port,0, &stop_char, 1);
        	#endif

		if (status <= 0) 
		{
			return;
		}
	}

	/* if we are implementing RTS/CTS, toggle that line */
	if (tty->termios->c_cflag & CRTSCTS) 
	{
		mos7720_port->shadowMCR &= ~MCR_RTS;
		status = SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,MCR, &mos7720_port->shadowMCR);
		if (status != 0) 
		{
			return;
		}
	}

	return;
}


/*****************************************************************************
 * mos7720_unthrottle
 *	this function is called by the tty driver when it wants to resume the data
 *	being read from the port (called after SerialThrottle is called)
 *****************************************************************************/
static void mos7720_unthrottle (struct usb_serial_port *port)
{
	struct tty_struct *tty;
	int status;
	struct moschip_port *mos7720_port = get_port_private(port); 


	if (mos7720_port == NULL)
		return;

	if (!mos7720_port->open) {
		dbg("%s - port not opened", __FUNCTION__);
		return;
	}

	DPRINTK("%s","Entering .......... \n");

	tty = port->tty;
	if (!tty) 
	{
		dbg ("%s - no tty available", __FUNCTION__);
		return;
	}

	/* if we are implementing XON/XOFF, send the start character */
	if (I_IXOFF(tty)) 
	{
		unsigned char start_char = START_CHAR(tty);
		#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
                status = mos7720_write (port, &start_char, 1); //FC4
	        #endif
	        #if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,10)
                status = mos7720_write (port,0, &start_char, 1);
        	#endif

		if (status <= 0) 
		{
			return;
		}
	}

	/* if we are implementing RTS/CTS, toggle that line */
	if (tty->termios->c_cflag & CRTSCTS) 
	{
		mos7720_port->shadowMCR |= MCR_RTS;
		status = SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,MCR, &mos7720_port->shadowMCR);
		if (status != 0) 
		{
			return;
		}
	}

	return;
}

static int mos7720_tiocmget(struct usb_serial_port *port, struct file *file)
{
        struct moschip_port *mos7720_port;
        unsigned int result;
        char msr;
        char mcr;
        mos7720_port = get_port_private(port);
                                                                                                                             
        DPRINTK("%s - port %d", __FUNCTION__, port->number);
                                                                                                                             
        if (mos7720_port == NULL)
                return -ENODEV;
                                                                                                                             
	SendMosCmd(port->serial,MOS_READ,port->number - port->serial->minor,MODEM_STATUS_REGISTER,&msr);
	SendMosCmd(port->serial,MOS_READ,port->number - port->serial->minor,MODEM_CONTROL_REGISTER,&mcr);
        result = ((mcr & MCR_DTR) ? TIOCM_DTR : 0)
                | ((mcr & MCR_RTS) ? TIOCM_RTS : 0)
                | ((mcr & MCR_LOOPBACK) ? TIOCM_LOOP : 0)
                | ((msr & MOS7720_MSR_CTS) ? TIOCM_CTS : 0)
                | ((msr & MOS7720_MSR_CD) ? TIOCM_CAR : 0)
                | ((msr & MOS7720_MSR_RI) ? TIOCM_RI : 0)
                | ((msr & MOS7720_MSR_DSR) ? TIOCM_DSR : 0);
                                                                                                                             
        DPRINTK("%s - 0x%04X", __FUNCTION__, result);
                                                                                                                             
        return result;
}

static int mos7720_tiocmset(struct usb_serial_port *port, struct file *file,
        unsigned int set, unsigned int clear)
{
        struct moschip_port *mos7720_port;
        unsigned int mcr;
        unsigned int status;
                                                                                                                             
        DPRINTK("%s - port %d", __FUNCTION__, port->number);
                                                                                                                             
        mos7720_port = get_port_private(port);
                                                                                                                             
        if (mos7720_port == NULL)
                return -ENODEV;
                                                                                                                             
                                                                                                                             
                                                                                                                             
         mcr = mos7720_port->shadowMCR;
        if (clear & TIOCM_RTS)
                mcr &= ~MCR_RTS;
        if (clear & TIOCM_DTR)
                mcr &= ~MCR_DTR;
        if (clear & TIOCM_LOOP)
                mcr &= ~MCR_LOOPBACK;
                                                                                                                             
        if (set & TIOCM_RTS)
                mcr |= MCR_RTS;
        if (set & TIOCM_DTR)
                mcr |= MCR_DTR;
        if (set & TIOCM_LOOP)
                mcr |= MCR_LOOPBACK;
                                                                                                                             
         mos7720_port->shadowMCR = mcr;
                                                                                                                             
        status=0;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,MODEM_STATUS_REGISTER,&mcr);
                                                                                                                             
        return 0;
}
                                                                                                                             




/*****************************************************************************
 * SerialSetTermios
 *	this function is called by the tty driver when it wants to change the termios structure
 *****************************************************************************/

static void mos7720_set_termios (struct usb_serial_port *port, struct termios *old_termios)
{
	int status;
	unsigned int cflag;
	struct usb_serial *serial;
	struct moschip_port *mos7720_port;
	struct tty_struct *tty;
	

	serial = port->serial;


	mos7720_port = get_port_private(port); 

	if (mos7720_port == NULL)
		return;

	tty = port->tty;

	if (!port->tty || !port->tty->termios)
	{
		dbg ("%s - no tty or termios", __FUNCTION__);
		return;
	}

	if (!mos7720_port->open) 
	{
		dbg("%s - port not opened", __FUNCTION__);
		return;
	}

	DPRINTK("%s\n","setting termios - ");

	cflag = tty->termios->c_cflag;

	if (!cflag)
	{
           printk("%s %s\n",__FUNCTION__,"cflag is NULL");
	   return;
	}            

	/* check that they really want us to change something */
	if (old_termios)
	{
		if ((cflag == old_termios->c_cflag) &&
		    (RELEVANT_IFLAG(tty->termios->c_iflag) == RELEVANT_IFLAG(old_termios->c_iflag))) 
		{
			DPRINTK("%s\n","Nothing to change");
			return;
		}
	}

	dbg("%s - clfag %08x iflag %08x", __FUNCTION__, 
	    tty->termios->c_cflag,
	    RELEVANT_IFLAG(tty->termios->c_iflag));
	
	if (old_termios) 
	{
		dbg("%s - old clfag %08x old iflag %08x", __FUNCTION__,
		    old_termios->c_cflag,
		    RELEVANT_IFLAG(old_termios->c_iflag));
	}

	dbg("%s - port %d", __FUNCTION__, port->number);

	/* change the port settings to the new ones specified */

	change_port_settings (mos7720_port, old_termios);

	if(!mos7720_port->read_urb)
	{
		DPRINTK("%s","URB KILLED !!!!!\n");
		return;
	}

	if(mos7720_port->read_urb->status!=-EINPROGRESS)
	{
		mos7720_port->read_urb->dev = serial->dev;
	 	status = usb_submit_urb(mos7720_port->read_urb, GFP_ATOMIC);
		if (status) 
		{
			DPRINTK(" usb_submit_urb(read bulk) failed, status = %d", status);
		}
	}
	return;
}


/*****************************************************************************
 * get_lsr_info - get line status register info
 *
 *****************************************************************************/

static int get_lsr_info(struct moschip_port *mos7720_port, unsigned int *value)
{
	int count;
	unsigned int result = 0;

        count = mos7720_chars_in_buffer(mos7720_port->port);
        if(count == 0)
	{
		dbg("%s -- Empty", __FUNCTION__);
		result = TIOCSER_TEMT;
	}

	if (copy_to_user(value, &result, sizeof(int)))
		return -EFAULT;
	return 0;
}

/*****************************************************************************
 * get_number_bytes_avail - get number of bytes available
 *
 * Purpose: Let user call ioctl to get the count of number of bytes available.
 *****************************************************************************/

static int get_number_bytes_avail(struct moschip_port *mos7720_port, unsigned int *value)
{
	unsigned int result = 0;
	struct tty_struct *tty = mos7720_port->port->tty;

	if (!tty)
		return -ENOIOCTLCMD;

	result = tty->read_cnt;

	dbg("%s(%d) = %d", __FUNCTION__,  mos7720_port->port->number, result);
	if (copy_to_user(value, &result, sizeof(int)))
		return -EFAULT;

	return -ENOIOCTLCMD;
}

/*****************************************************************************
 * set_modem_info
 *      function to set modem info
 *****************************************************************************/

static int set_modem_info(struct moschip_port *mos7720_port, unsigned int cmd, unsigned int *value)
{
	unsigned int mcr ;
	unsigned int arg;
	unsigned char data;

	struct usb_serial_port *port;

	if (mos7720_port == NULL)
		return -1;


	port = (struct usb_serial_port*)mos7720_port->port;

	mcr = mos7720_port->shadowMCR;

	if (copy_from_user(&arg, value, sizeof(int)))
		return -EFAULT;

	switch (cmd) {
		case TIOCMBIS:
			if (arg & TIOCM_RTS)
				mcr |= MCR_RTS;
			if (arg & TIOCM_DTR)
				mcr |= MCR_RTS;
			if (arg & TIOCM_LOOP)
				mcr |= MCR_LOOPBACK;
			break;

		case TIOCMBIC:
			if (arg & TIOCM_RTS)
				mcr &= ~MCR_RTS;
			if (arg & TIOCM_DTR)
				mcr &= ~MCR_RTS;
			if (arg & TIOCM_LOOP)
				mcr &= ~MCR_LOOPBACK;
			break;

		case TIOCMSET:
			/* turn off the RTS and DTR and LOOPBACK 
			 * and then only turn on what was asked to */
			mcr &=  ~(MCR_RTS | MCR_DTR | MCR_LOOPBACK);
			mcr |= ((arg & TIOCM_RTS) ? MCR_RTS : 0);
			mcr |= ((arg & TIOCM_DTR) ? MCR_DTR : 0);
			mcr |= ((arg & TIOCM_LOOP) ? MCR_LOOPBACK : 0);
			break;
	}

	mos7720_port->shadowMCR = mcr;

	data = mos7720_port->shadowMCR;
	SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,MCR,&data);

	return 0;
}

/*****************************************************************************
 * get_modem_info
 *      function to get modem info
 *****************************************************************************/

static int get_modem_info(struct moschip_port *mos7720_port, unsigned int *value)
{
	unsigned int result = 0;
	unsigned int msr = mos7720_port->shadowMSR;
	unsigned int mcr = mos7720_port->shadowMCR;

	result = ((mcr & MCR_DTR)	? TIOCM_DTR: 0)	  /* 0x002 */
		  | ((mcr & MCR_RTS)	? TIOCM_RTS: 0)   /* 0x004 */
		  | ((msr & MOS7720_MSR_CTS)	? TIOCM_CTS: 0)   /* 0x020 */
		  | ((msr & MOS7720_MSR_CD)	? TIOCM_CAR: 0)   /* 0x040 */
		  | ((msr & MOS7720_MSR_RI)	? TIOCM_RI:  0)   /* 0x080 */
		  | ((msr & MOS7720_MSR_DSR)	? TIOCM_DSR: 0);  /* 0x100 */


	dbg("%s -- %x", __FUNCTION__, result);

	if (copy_to_user(value, &result, sizeof(int)))
		return -EFAULT;
	return 0;
}

/*****************************************************************************
 * get_serial_info
 *      function to get information about serial port
 *****************************************************************************/

static int get_serial_info(struct moschip_port *mos7720_port, struct serial_struct * retinfo)
{
	struct serial_struct tmp;

	if (mos7720_port == NULL)
		return -1;


	if (!retinfo)
		return -EFAULT;

	memset(&tmp, 0, sizeof(tmp));

	tmp.type		= PORT_16550A;
	tmp.line		= mos7720_port->port->serial->minor;
	tmp.port		= mos7720_port->port->number;
	tmp.irq			= 0;
	tmp.flags		= ASYNC_SKIP_TEST | ASYNC_AUTO_IRQ;
        tmp.xmit_fifo_size      = NUM_URBS * URB_TRANSFER_BUFFER_SIZE;
	tmp.baud_base		= 9600;
	tmp.close_delay		= 5*HZ;
	tmp.closing_wait	= 30*HZ;


	if (copy_to_user(retinfo, &tmp, sizeof(*retinfo)))
		return -EFAULT;
	return 0;
}

/*****************************************************************************
 * SerialIoctl
 *	this function handles any ioctl calls to the driver
 *****************************************************************************/

static int mos7720_ioctl (struct usb_serial_port *port, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct moschip_port *mos7720_port;
	//struct tty_struct *tty;
	struct async_icount cnow;
	struct async_icount cprev;
	struct serial_icounter_struct icount;
	

	mos7720_port = get_port_private(port); 
	if (mos7720_port == NULL)
		return -1;

	dbg("%s - port %d, cmd = 0x%x", __FUNCTION__, port->number, cmd);

	switch (cmd) 
	{
                /* return number of bytes available */
	
		case TIOCINQ:
			dbg("%s (%d) TIOCINQ", __FUNCTION__,  port->number);
			return get_number_bytes_avail(mos7720_port, (unsigned int *) arg);
			break;
/*
		case TIOCOUTQ:
                        dbg("%s (%d) TIOCOUTQ", __FUNCTION__,  port->number);
                        return put_user(tty->driver->chars_in_buffer ?
                                        tty->driver->chars_in_buffer(tty) : 0,
                                        (int __user *) arg);
                        break;
*/
                                                                                                                             
                /*  //2.6.17 block
                case TCFLSH:
                        retval = tty_check_change(tty);
                        if (retval)
                                return retval;
                                                                                                                             
                        ld = tty_ldisc_ref(tty);
                        switch (arg) {
                                case TCIFLUSH:
                                        if (ld && ld->flush_buffer)
                                                ld->flush_buffer(tty);
                                        break;
                                case TCIOFLUSH:
                                        if (ld && ld->flush_buffer)
                                                ld->flush_buffer(tty);
                                        // fall through
                                case TCOFLUSH:
                                        if (tty->driver->flush_buffer)
                                                tty->driver->flush_buffer(tty);
                                        break;
                                default:
                                        tty_ldisc_deref(ld);
                                        return -EINVAL;
                        }
                        tty_ldisc_deref(ld);
                        return 0;
                */
/*
		case TCGETS:
                        if (kernel_termios_to_user_termios((struct termios __user *)arg, tty->termios))
                                return -EFAULT;
                        return 0;
*/		

		case TIOCSERGETLSR:
			dbg("%s (%d) TIOCSERGETLSR", __FUNCTION__,  port->number);
			return get_lsr_info(mos7720_port, (unsigned int *) arg);
			return 0;

		case TIOCMBIS:
		case TIOCMBIC:
		case TIOCMSET:
			dbg("%s (%d) TIOCMSET/TIOCMBIC/TIOCMSET", __FUNCTION__,  port->number);
			return set_modem_info(mos7720_port, cmd, (unsigned int *) arg);

		case TIOCMGET:  
			dbg("%s (%d) TIOCMGET", __FUNCTION__,  port->number);
			return get_modem_info(mos7720_port, (unsigned int *) arg);

		case TIOCGSERIAL:
			dbg("%s (%d) TIOCGSERIAL", __FUNCTION__,  port->number);
			return get_serial_info(mos7720_port, (struct serial_struct *) arg);

		case TIOCSSERIAL:
			dbg("%s (%d) TIOCSSERIAL", __FUNCTION__,  port->number);
			break;

		case TIOCMIWAIT:
			dbg("%s (%d) TIOCMIWAIT", __FUNCTION__,  port->number);
			cprev = mos7720_port->icount;
			while (1) {
				//interruptible_sleep_on(&mos7720_port->delta_msr_wait);
				/* see if a signal did it */
				if (signal_pending(current))
					return -ERESTARTSYS;
				cnow = mos7720_port->icount;
				if (cnow.rng == cprev.rng && cnow.dsr == cprev.dsr &&
				    cnow.dcd == cprev.dcd && cnow.cts == cprev.cts)
					return -EIO; /* no change => error */
				if (((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
				    ((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
				    ((arg & TIOCM_CD)  && (cnow.dcd != cprev.dcd)) ||
				    ((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) ) {
					return 0;
				}
				cprev = cnow;
			}
			/* NOTREACHED */
			break;

		case TIOCGICOUNT:
			cnow = mos7720_port->icount;
			icount.cts = cnow.cts;
			icount.dsr = cnow.dsr;
			icount.rng = cnow.rng;
			icount.dcd = cnow.dcd;
			icount.rx = cnow.rx;
			icount.tx = cnow.tx;
			icount.frame = cnow.frame;
			icount.overrun = cnow.overrun;
			icount.parity = cnow.parity;
			icount.brk = cnow.brk;
			icount.buf_overrun = cnow.buf_overrun;

			dbg("%s (%d) TIOCGICOUNT RX=%d, TX=%d", __FUNCTION__,  port->number, icount.rx, icount.tx );
			if (copy_to_user((void *)arg, &icount, sizeof(icount)))
				return -EFAULT;
			return 0;

		case TIOCEXBAUD:
			return 0;
	}

	return -ENOIOCTLCMD;
}





/*****************************************************************************
 * mos7720_send_cmd_write_baud_rate
 *      this function sends the proper command to change the baud rate of the
 *      specified port.
 *****************************************************************************/

static int mos7720_send_cmd_write_baud_rate (struct moschip_port *mos7720_port, int baudRate)
{
        int divisor = 0;
        int status;
        //__u16 Data;
        unsigned char Data;
        unsigned char data;
        unsigned char number ;
        __u16 clk_sel_val;
        struct usb_serial_port *port;

        if (mos7720_port == NULL)
                return -1;

        port = (struct usb_serial_port*)mos7720_port->port;


        DPRINTK("%s","Entering .......... \n");

        number = mos7720_port->port->number - mos7720_port->port->serial->minor;

        dbg("%s - port = %d, baud = %d", __FUNCTION__, mos7720_port->port->number, baudRate);
        //reset clk_uart_sel in spregOffset

        clk_sel_val=0x0;
        Data=0x0;
        status=0;
        status = mos7720_calc_baud_rate_divisor (baudRate, &divisor,&clk_sel_val);
       	SendMosCmd(mos7720_port->port->serial,MOS_READ,MOS_MAX_PORT,(port->number - port->serial->minor)+1,&Data);
        Data = (Data & 0x8f)|clk_sel_val;
       	SendMosCmd(mos7720_port->port->serial,MOS_WRITE,MOS_MAX_PORT,(port->number - port->serial->minor)+1,&Data);
                 /* Calculate the Divisor */

	/* Enable access to divisor latch */
        data = mos7720_port->shadowLCR | LCR_DL_ENABLE;
        mos7720_port->shadowLCR  = data;
        SendMosCmd(mos7720_port->port->serial,MOS_WRITE,port->number - port->serial->minor,LCR, &data);

        /* Write the divisor */
        data = LOW8 (divisor);
        SendMosCmd(mos7720_port->port->serial,MOS_WRITE,port->number - port->serial->minor,0x00, &data);

        data = HIGH8 (divisor);
        SendMosCmd(mos7720_port->port->serial,MOS_WRITE,port->number - port->serial->minor,0x01, &data);

        /* Disable access to divisor latch */
        data = mos7720_port->shadowLCR & ~LCR_DL_ENABLE;
        mos7720_port->shadowLCR = data;
        SendMosCmd(mos7720_port->port->serial,MOS_WRITE,port->number - port->serial->minor,0x03, &data);
	
		
        

        return status;
}


/*****************************************************************************
 * mos7720_calc_baud_rate_divisor
 *      this function calculates the proper baud rate divisor for the specified
 *      baud rate.
 *****************************************************************************/
static int mos7720_calc_baud_rate_divisor (int baudRate, int *divisor,__u16 *clk_sel_val)
{
        //int i;
        //__u16 custom,round1, round;

        dbg("%s - %d", __FUNCTION__, baudRate);

                if(baudRate <=115200)
                {
                        *divisor = 115200/baudRate;
                        *clk_sel_val = 0x0;
                }
                if((baudRate > 115200) && (baudRate <= 230400))
                {
                        *divisor = 230400/baudRate;
                        *clk_sel_val=0x10;
                }
                else if((baudRate > 230400) && (baudRate <= 403200))
                {
                        *divisor = 403200/baudRate;
                        *clk_sel_val=0x20;
                }
                else if((baudRate > 403200) && (baudRate <= 460800))
                {
                        *divisor = 460800/baudRate;
                        *clk_sel_val=0x30;
                }
                else if((baudRate > 460800) && (baudRate <= 806400))
                {
                        *divisor = 806400/baudRate;
                        *clk_sel_val=0x40;
                }
                else if((baudRate >806400) && (baudRate <= 921600))
                {
                        *divisor = 921600/baudRate;
                        *clk_sel_val=0x50;
                }
        return 0;

}






/*****************************************************************************
 * change_port_settings
 *	This routine is called to set the UART on the device to match 
 *      the specified new settings.
 *****************************************************************************/

static void change_port_settings (struct moschip_port *mos7720_port, struct termios *old_termios)
{
	struct tty_struct *tty;
	int baud;
	unsigned cflag;
	unsigned iflag;
	__u8 mask = 0xff;
	__u8 lData;
	__u8 lParity;
	__u8 lStop;
	int status;
	char data;

	struct usb_serial_port *port;
	struct usb_serial *serial;

	if (mos7720_port == NULL)
		return ;

	port = (struct usb_serial_port *)mos7720_port->port;



	serial = port->serial;

	dbg("%s - port %d", __FUNCTION__, mos7720_port->port->number);

	if ((!mos7720_port->open) && (!mos7720_port->openPending)) 
	{
		dbg("%s - port not opened", __FUNCTION__);
		return;
	}

	tty = mos7720_port->port->tty;
	
	if ((!tty) || (!tty->termios)) 
	{
		dbg("%s - no tty structures", __FUNCTION__);
		return;
	}

	DPRINTK("%s","Entering .......... \n");

	lData = LCR_BITS_8;
	lStop = LCR_STOP_1;         
	lParity = LCR_PAR_NONE;

	cflag = tty->termios->c_cflag;
	iflag = tty->termios->c_iflag;

	/* Change the number of bits */	
	switch (cflag & CSIZE) 
	{
		case CS5:	lData = LCR_BITS_5; 
				mask = 0x1f;    
			    	break;

		case CS6:   	lData = LCR_BITS_6; 
				mask = 0x3f;    
			    	break;

		case CS7:   	lData = LCR_BITS_7; 
				mask = 0x7f;    
			    	break;
		default:
		case CS8:   	lData = LCR_BITS_8;
			    	break;
	}

	/* Change the Parity bit */
	if (cflag & PARENB) 
	{
		if (cflag & PARODD) 
		{
			lParity = LCR_PAR_ODD;
			dbg("%s - parity = odd", __FUNCTION__);
		} 
		else 
		{
			lParity = LCR_PAR_EVEN;
			dbg("%s - parity = even", __FUNCTION__);
		}

	} 
	else 
	{
		dbg("%s - parity = none", __FUNCTION__);
	}
	
	if(cflag & CMSPAR)
	{
		lParity = lParity | 0x20;
	}

	/* Change the Stop bit */
	if (cflag & CSTOPB) 
	{
		lStop = LCR_STOP_2;
		dbg("%s - stop bits = 2", __FUNCTION__);
	} 
	else 
	{
		lStop = LCR_STOP_1;
		dbg("%s - stop bits = 1", __FUNCTION__);
	}


	/* Update the LCR with the correct value */
	mos7720_port->shadowLCR &= ~(LCR_BITS_MASK | LCR_STOP_MASK | LCR_PAR_MASK);
	mos7720_port->shadowLCR |= (lData | lParity | lStop);

	mos7720_port->validDataMask = mask;

	/* Disable Interrupts */
	data = 0x00;
        SendMosCmd(mos7720_port->port->serial,MOS_WRITE,port->number - port->serial->minor,IER, &data);

	data = 0x00;
        SendMosCmd(mos7720_port->port->serial,MOS_WRITE,port->number - port->serial->minor,FCR, &data);

	data = 0xcf;
        SendMosCmd(mos7720_port->port->serial,MOS_WRITE,port->number - port->serial->minor,FCR, &data);

	/* Send the updated LCR value to the mos7720 */
	data = mos7720_port->shadowLCR;
        SendMosCmd(mos7720_port->port->serial,MOS_WRITE,port->number - port->serial->minor,LCR, &data);

        data = 0x00b;
        mos7720_port->shadowMCR = data;
        SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x04, &data);
        data = 0x00b;
        SendMosCmd(port->serial,MOS_WRITE,port->number - port->serial->minor,0x04, &data);

	/* set up the MCR register and send it to the mos7720 */
	
	mos7720_port->shadowMCR = MCR_MASTER_IE;
	if (cflag & CBAUD) 
	{
		mos7720_port->shadowMCR |= (MCR_DTR | MCR_RTS);
	}


	if (cflag & CRTSCTS) 
	{
		mos7720_port->shadowMCR |= (MCR_XON_ANY);

                /* To set hardware flow control to the specified *
                 * serial port, in SP1/2_CONTROL_REG             */

		if(port->number)
		{
		  data = 0x001;
		  SendMosCmd(port->serial,MOS_WRITE,MOS_MAX_PORT,0x08, &data);
		}
		else
		{
		  data = 0x002;
		  SendMosCmd(port->serial,MOS_WRITE,MOS_MAX_PORT,0x08, &data);
		}
	}
	else
	{
		mos7720_port->shadowMCR &= ~(MCR_XON_ANY); 
	}


	data = mos7720_port->shadowMCR;
	SendMosCmd(mos7720_port->port->serial,MOS_WRITE,port->number - port->serial->minor,MCR, &data);
	


	/* Determine divisor based on baud rate */
	baud = tty_get_baud_rate(tty);

	dbg("%s - baud rate = %d", __FUNCTION__, baud);
	printk("%s - baud rate = %d", __FUNCTION__, baud);

	if (!baud) 
	{
		/* pick a default, any default... */
        	DPRINTK("%s\n","Picked default baud...");
		baud = 9600;
	}

	/*
	if(baud >= 230400)
	{
		set_higher_rates(mos7720_port,&baud);
	        // Enable Interrupts 
        	data = 0x0c;
        	SendMosCmd(mos7720_port->port->serial,MOS_WRITE,port->number - port->serial->minor,IER, &data);
		return;
	}	  	

	*/
	status = mos7720_send_cmd_write_baud_rate (mos7720_port, baud);




	/* Enable Interrupts */
	data = 0x0c;
        SendMosCmd(mos7720_port->port->serial,MOS_WRITE,port->number - port->serial->minor,IER, &data);

	if(mos7720_port->read_urb->status!=-EINPROGRESS)
	{
		mos7720_port->read_urb->dev = serial->dev;
		
		status = usb_submit_urb(mos7720_port->read_urb, GFP_ATOMIC);
	
		if (status) 
		{
			DPRINTK(" usb_submit_urb(read bulk) failed, status = %d", status);
		}
	}
	return;
}

/****************************************************************************
 * mos7720_startup
 ****************************************************************************/

static int mos7720_startup (struct usb_serial *serial)
{
	struct moschip_serial *mos7720_serial;
	struct moschip_port *mos7720_port;
	struct usb_device *dev;
	int i;
	char data;

//	DPRINTK("%s \n"," mos7720_startup :entering..........");

	if(!serial)
	{
		DPRINTK("%s\n","Invalid Handler");
		return -1;
	}

	dev = serial->dev;
	
	DPRINTK("%s\n","Entering...");

	/* create our private serial structure */
	mos7720_serial = kmalloc (sizeof(struct moschip_serial), GFP_KERNEL);
	if (mos7720_serial == NULL) 
	{
		err("%s - Out of memory", __FUNCTION__);
		return -ENOMEM;
	}

	/* resetting the private structure field values to zero */
	memset (mos7720_serial, 0, sizeof(struct moschip_serial));

	mos7720_serial->serial = serial;
	set_serial_private(serial,mos7720_serial);

	/* we set up the pointers to the endpoints in the mos7720_open *
	 * function, as the structures aren't created yet.             */

	/* set up port private structures */
	for (i = 0; i < serial->num_ports; ++i) 
	{

		DPRINTK("port number is %d \n",serial->port[i]->number);
		DPRINTK("serial number is %d \n",serial->minor);
	        mos7720_port = kmalloc(sizeof(struct moschip_port), GFP_KERNEL);
		if (mos7720_port == NULL) 
		{
			err("%s - Out of memory", __FUNCTION__);
			set_serial_private(serial,NULL);
			kfree(mos7720_serial);
			return -ENOMEM;
		}
		memset (mos7720_port, 0, sizeof(struct moschip_port));


	/* Initialize all port interrupt end point to port 0 int endpoint *
	 * Our device has only one interrupt end point comman to all port */

		//serial->port[i]->interrupt_in_endpointAddress = serial->port[0]->interrupt_in_endpointAddress;

		mos7720_port->port = serial->port[i];
		set_port_private(serial->port[i],mos7720_port);

		mos7720_port->control_urb = usb_alloc_urb(0,SLAB_ATOMIC);
        	mos7720_port->ctrl_buf = kmalloc(16,GFP_KERNEL);

	}


	/* setting configuration feature to one */
	usb_control_msg (serial->dev, usb_sndctrlpipe(serial->dev, 0), (__u8)0x03, 0x00,0x01,0x00, 0x00, 0x00, 5*HZ);

	
        SendMosCmd(serial,MOS_READ,0x00,LSR, &data);  // LSR For Port 1
	DPRINTK("LSR:%x\n",data);
        
	SendMosCmd(serial,MOS_READ,0x01,LSR, &data);  // LSR For Port 2
	DPRINTK("LSR:%x\n",data);

/*      MATRIX  */

	return 0;
}



/****************************************************************************
 * mos7720_shutdown
 *	This function is called whenever the device is removed from the usb bus.
 ****************************************************************************/

static void mos7720_shutdown (struct usb_serial *serial)
{
	int i;
        struct moschip_port *mos7720_port;


	DPRINTK("%s \n"," shutdown :entering..........");

/* MATRIX  */
	//ThreadState = 1;
/* MATRIX  */

	if(!serial)
	{
		DPRINTK("%s","Invalid Handler \n");
		return;
	}

	/*	check for the ports to be closed,close the ports and disconnect		*/

	/* free private structure allocated for serial port  * 
	 * stop reads and writes on all ports   	     */

	for (i=0; i < serial->num_ports; ++i) 
	{
		mos7720_port = get_port_private(serial->port[i]);
                kfree(mos7720_port->ctrl_buf);
        #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
                usb_kill_urb(mos7720_port->control_urb);
        #endif
        #if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,10)
                usb_unlink_urb(mos7720_port->control_urb);
        #endif
                kfree(mos7720_port);
                set_port_private(serial->port[i],NULL);	

	}

	/* free private structure allocated for serial device */

	kfree(get_serial_private(serial));
	set_serial_private(serial,NULL);

        DPRINTK("%s\n","Thank u ");

}


/****************************************************************************
 * moschip7720_init
 *	This is called by the module subsystem, or on startup to initialize us
 ****************************************************************************/
 int __init moschip7720_init(void)
{
	int retval;

	DPRINTK("%s \n"," mos7720_init :entering..........");

        /* Register with the usb serial */
	retval = usb_serial_register (&moschip7720_2port_device);

	if(retval)
		goto failed_port_device_register;

	DPRINTK("%s\n","Entring...");
	info(DRIVER_DESC " " DRIVER_VERSION);


 	/* Register with the usb */
	retval = usb_register(&io_driver);

	if (retval) 
		goto failed_usb_register;

	if(retval == 0) 
	{
		DPRINTK("%s\n","Leaving...");
		return 0;
	}


failed_usb_register:
	usb_serial_deregister(&moschip7720_2port_device);

failed_port_device_register:

	return retval;
}

/****************************************************************************
 * moschip7720_exit
 *	Called when the driver is about to be unloaded.
 ****************************************************************************/
void __exit moschip7720_exit (void)
{

	DPRINTK("%s \n"," mos7720_exit :entering..........");

	usb_deregister (&io_driver);

	usb_serial_deregister (&moschip7720_2port_device);

	DPRINTK("%s\n","Entring...");
}

module_init(moschip7720_init);
module_exit(moschip7720_exit);

/* Module information */
MODULE_DESCRIPTION( DRIVER_DESC );
MODULE_LICENSE("GPL");

MODULE_PARM_DESC(debug, "Debug enabled or not");

